"""The main entry for bob.vuln
"""
import click
import pkg_resources

from clapper.click import AliasedGroup
from click_plugins import with_plugins


@with_plugins(pkg_resources.iter_entry_points("bob.vuln.cli"))
@click.group(cls=AliasedGroup)
def vulnerability():
    """Vulnerability analysis related commands."""
    pass
