from bob.bio.base.config.dummy.algorithm import algorithm  # noqa: F401
from bob.bio.base.config.dummy.database import database  # noqa: F401
from bob.bio.base.config.dummy.extractor import extractor  # noqa: F401
from bob.bio.base.config.dummy.preprocessor import preprocessor  # noqa: F401

zt_norm = True
verbose = 1
sub_directory = "test_dummy"
